#include <Engine/Globals.h>
#include <Engine/types/Scene.h>
#include <Engine/types/entity/Camera.h>
#include <Engine/types/entity/InstancedSprite.h>
#include <Engine/Logger.h>
#include <JGL/JGL.h>
#include <algorithm>
#include <thread>

using namespace Engine;
size_t Scene::EntityCount() const {
    return Entity::GetAllDescendents(entity_list).size();
}

void Scene::Update() {
    auto target_tick_time = std::chrono::microseconds(1000000 / tick_rate);

    auto start = std::chrono::high_resolution_clock::now();

    entity_list->UpdateChildren();

    auto stop = std::chrono::high_resolution_clock::now();
    auto elapsed = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);

    if (elapsed < target_tick_time)
        std::this_thread::sleep_for(std::chrono::microseconds(target_tick_time - elapsed));
}

void Scene::UpdateLoop() {
    while (!destroying)
        if (should_tick)
            Update();
        else {
            auto target_tick_time = std::chrono::microseconds(1000000 / tick_rate);
            std::this_thread::sleep_for(std::chrono::microseconds(target_tick_time));
        }
}

void Scene::Render() {
    std::vector<Renderable*> display_list{};
    for (auto *e: Entity::GetAllDescendents(entity_list))
        if (auto *r = dynamic_cast<Renderable *>(e))
            if (r->ShouldRender())
                display_list.push_back(r);

    std::sort(display_list.begin(), display_list.end(), [](Renderable *a, Renderable *b) {
        return a->GetLayer() > b->GetLayer();
    });

    if (active_camera)
        active_camera->Render();

    for (auto &r: display_list)
        if (r->ShouldRender())
            r->Render();
    J2D::End();
}

void Scene::Render(JGL::RenderTarget* render_target) {
    if (!render_target)
        Logger::Fatal("Rendering the current scene to a RenderTarget that is nullptr?");

    J2D::Begin(render_target, true);
        Render();
    J2D::End();
    last_rendered_to_window = false;
}

void Scene::Render(ReWindow::RWindow* window) {
    if (!window)
        Logger::Fatal("Rendering the current scene to a ReWindow that is nullptr?");

    J2D::Begin(nullptr, true);
        Render();
    J2D::End();
    last_rendered_to_window = true;
}

Scene::~Scene() {
    Scene::Unload();
}

void Scene::Unload() {
    destroying = true;
    while (!update_worker->Done()) {}
    delete update_worker;

    for (auto* itx : instanced_textures)
        delete itx;

    for (auto* a : instanced_alpha_masks)
        delete a;

    delete entity_list;
    delete active_camera;
}

std::string Scene::GetName() const {
    return name;
}

Camera* Scene::GetActiveCamera() const {
    return active_camera;
}

InstancedTexture* Scene::GetInstancedTexture(const InstancedSprite* user) {
    for (auto* itx : instanced_textures)
        if (itx->InUseBy(user))
            return itx;

    auto* t = new Texture(user->GetTextureFilesystemPath());
    auto* itx = new InstancedTexture(t, user);
    instanced_textures.push_back(itx);
    return itx;
}

InstancedTexture* Scene::GetInstancedAlphaMask(const InstancedSprite* user) {
    for (auto* itx : instanced_alpha_masks)
        if (itx->InUseBy(user))
            return itx;

    if (!user->GetAlphaMaskFilesystemPath().empty()) {
        auto *t = new Texture(user->GetAlphaMaskFilesystemPath());
        auto *itx = new InstancedTexture(t, user);
        instanced_alpha_masks.push_back(itx);
        return itx;
    }
    return nullptr;
}

void Scene::DestroyInstancedTexture(const InstancedTexture* itx) {
    auto it = std::find(instanced_textures.begin(), instanced_textures.end(), itx);
    if (it != instanced_textures.end())
        delete *it, instanced_textures.erase(it);
}

void Scene::DestroyInstancedAlphaMask(const Engine::InstancedTexture* alpha_mask) {
    auto it = std::find(instanced_alpha_masks.begin(), instanced_alpha_masks.end(), alpha_mask);
    if (it != instanced_alpha_masks.end())
        delete *it, instanced_alpha_masks.erase(it);
}

void Scene::Run() {
    update_worker = new ConcurrentWorker([this] { UpdateLoop(); });
}

bool Scene::ShouldTick() const {
    return should_tick;
}

void Scene::ShouldTick(bool state) {
    should_tick = state;
}

float Scene::TickDeltaTime() const {
    return (1000000.f / tick_rate) / 1000000.f;
}

float Scene::FrameDeltaTime() const {
    if (last_rendered_to_window && Globals::Window)
        return Globals::Window->GetDeltaTime();
    return render_target_frame_delta;
}

